/* net-probe.c */

/* Copyright (C) 2002 Niels Mller
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <assert.h>
#include <ctype.h>
#include <errno.h>
#include <math.h>
#include <stdarg.h>
#include <stddef.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <time.h>

#include <unistd.h>

#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>

#include <fcntl.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>

#ifndef SOL_IP
# define SOL_IP IPPROTO_IP
#endif

#ifdef IP_RECVERR
# include <sys/uio.h>
# include <netinet/ip_icmp.h>

/* Should be in sys/errqueue.h */
struct sock_extended_err
  {
    u_int32_t ee_errno;
    u_int8_t ee_origin;
    u_int8_t ee_type;
    u_int8_t ee_code;
    u_int8_t ee_pad;
    u_int32_t ee_info;
    u_int32_t ee_data;
  };

#define SO_EE_ORIGIN_NONE  0
#define SO_EE_ORIGIN_LOCAL 1
#define SO_EE_ORIGIN_ICMP  2
#define SO_EE_ORIGIN_ICMP6 3

#define SO_EE_OFFENDER(see)	\
  ((struct sockaddr *)((see)+1))

#endif /* IP_RECVERR */

/* Use Dan Egnor's liboop */
#include <oop.h>

#include "echo.h"
#include "statistics.h"

#define DEFAULT_TTL 50

/* Number of seconds to wait for the network to return all our packets. */
#define DEFAULT_TIMEOUT 10

struct packet_info
{
  enum { SENT, SEND_ERROR, RECEIVED, DUPLICATE, INVALID, LOST } state;
  int err; /* errno if sending failed */
  struct timeval time_sent;
  struct timeval time_received;
  int ttl;
};

/* Represents one node, and the *incoming* link to that node. When
 * measuring a four-hop forward path, the nodes and links
 * corresponding to hop number i are:
 *
 * (0) -> |1| -> |2| -> |3| -> (4)
 *      1      2      3      4
 *
 * But when measuring the reverse path, we have
 *
 * (0) <- |1| <- |2| <- |3| <- (4)
 *     0      1      2      3
 */
struct hop_info
{
  /* Bandwidth to this hop, 0.0 if not known. */
  double bandwidth;
  double bandwidth_dev;

  /* Address of node, or INADDR_ANY if not known. */
  struct in_addr addr;

  const char *name;
};

struct error_info
{
  struct in_addr where;
  enum error_type { ERR_TIME_EXCEEDED = 1, ERR_OTHER = 0 } type;
};


struct packet_state;

/* Returns 1 when done .*/
typedef int
job_function(void *ctx, struct packet_state *state);

struct packet_state
{
  unsigned to_recv; /* Number of outstanding packets. */

  /* For cancelling the timeout */
  struct timeval timeout;
  
  unsigned long interval; /* Micro seconds */
  job_function *current;
  void *ctx;
  
  struct in_addr local;
  struct in_addr target;
  int socket;

  /* Statistics */
  /* Bandwidth estimate, from unloaded network */
  double bandwidth_fast;
  double bandwidth_fast_dev; /* std dev / mean */

  /* Minimum observed rtt for unloaded network. */
  double rtt_small;
  double rtt_small_dev;      /* std dev / mean */

  /* The number of hops to the router on the other side of the
   * bottleneck. */
  unsigned bottleneck;

  /* rtt when the network is loaded, and the queue before the bottle
   * neck link is full. */
  double rtt_large;

  /* Bandwidth, measured when the network is loaded. */
  double bandwidth;

  /* Buffer size estimated from the drops */
  double buffer_drops;

  /* Buffer size estiamted from the delays */
  double buffer_delay;

  /* Per hop information. Indexed by ttl, from 0 to nhops. 0
   * corresponds to the local machine, and nhops to the target
   * machine. */
  unsigned nhops;
  unsigned hop_alloc;
  struct hop_info *hops;
  
  /* Allocated packets */
  unsigned packet_alloc;
  /* Number of packets in the array */
  unsigned npackets;
  struct packet_info *packets;

  /* Received errors */
  unsigned error_alloc;
  unsigned nerrors;
  struct error_info *errors;
};

/* Packet sent on the wire. */
struct packet_contents
{
  char header[4];
  struct timeval stamp;
  unsigned i;
  char data[800];
};

#define CONTENTS_DATA_OFFSET (offsetof(struct packet_contents, data))
#define CONTENTS_ECHO_SMALL \
  ( (CONTENTS_DATA_OFFSET + ECHO_SIZE_INCREMENT - 1) / ECHO_SIZE_INCREMENT)

#define CONTENTS_ECHO_LARGE \
  ( (sizeof(struct packet_contents) - 1) / ECHO_SIZE_INCREMENT)

#define CONTENTS_SIZE_SMALL (CONTENTS_ECHO_SMALL * ECHO_SIZE_INCREMENT)
#define CONTENTS_SIZE_LARGE (CONTENTS_ECHO_LARGE * ECHO_SIZE_INCREMENT)

/* Packet size in bits */
#define PACKET_SIZE (8 * CONTENTS_SIZE_LARGE)

struct options
{
  const char *target;
  int port;
  int full;
  int long_report;
  int reverse;
};

static int quiet_flag = 0;
static int verbose_flag = 0;


static void
werror(const char *format, ...)
     __attribute__((__format__ (__printf__,1, 2)));

static void
verbose(const char *format, ...)
     __attribute__((__format__ (__printf__,1, 2)));

#define N(a) (sizeof(a) / sizeof(*(a)))

static void
print_pretty(FILE *f, double x, const char *unit)
{
  if (x < 0)
    {
      /* Abnormal case */
      x = -x;
      fprintf(f, "-");
    }
  if (x >= 1.0)
    {
      unsigned magnitude;
      const char *prefix[] =
	{ "", "K", "M", "G" };
      for (magnitude = 0;
	   magnitude + 1 < N(prefix) && x >= 1024.0;
	   magnitude++)
	x /= 1024.0;

      fprintf(f, "%.1f %s%s", x, prefix[magnitude], unit);
    }
  else
    {
      unsigned magnitude;
      const char *prefix[] =
	{ "", "m", "u" };
      for (magnitude = 0;
	   magnitude + 1 < N(prefix) && x < 1.0;
	   magnitude++)
	x *= 1000.0;

      fprintf(f, "%.1f %s%s", x, prefix[magnitude], unit);
    }
}

#undef N

/* A hack, using a format string of exactly %f%s invokes
 * the pretty printer. */
static int
vprint_pretty(FILE *f, const char *format, va_list args)
{
  if (!strcmp("%f%s", format))
    {
      double value = va_arg(args, double);
      const char *unit = va_arg(args, const char *);
      print_pretty(f, value, unit);
      return 1;
    }
  else
    return 0;
}

static void
werror(const char *format, ...)
     __attribute__((__format__ (__printf__,1, 2)));

static void
werror(const char *format, ...)
{
  if (!quiet_flag)
    {
      va_list args;
      va_start(args, format);
      if (vprint_pretty(stderr, format, args) == 0)
	vfprintf(stderr, format, args);
      va_end(args);
    }
}

static void
verbose(const char *format, ...)
{
  if (verbose_flag)
    {
      va_list args;
      va_start(args, format);
      if (vprint_pretty(stderr, format, args) == 0)
	vfprintf(stderr, format, args);
      va_end(args);
    }
}

static char *
sprintf_alloc(const char *format, ...)
{
  size_t alloc;
  char *s = NULL;

  for (alloc = 10;; alloc *= 2)
    {
      va_list args;
      int res;
      
      s = realloc(s, alloc);
      if (!s)
	{
	  werror("Virtual memory exhausted.\n");
	  abort();
	}

      va_start(args, format);
      res = vsnprintf(s, alloc, format, args);
      va_end(args);

      if (res < alloc)
	return s;
    }
}

static void
packet_state_init(struct packet_state *state)
{
  state->to_recv = 0;
  state->socket = 0;

  state->hop_alloc = 0;
  state->hops = NULL;
  state->nhops = 0;

  state->packet_alloc = 0;
  state->packets = NULL;
  state->npackets = 0;

  state->error_alloc = 0;
  state->errors = NULL;
  state->nerrors = 0;
}

static void
packet_state_alloc(struct packet_state *state, unsigned n)
{
  if (state->npackets + n > state->packet_alloc)
    {
      state->packet_alloc = state->npackets + n;
      state->packets = realloc(state->packets,
			       state->packet_alloc * sizeof(*state->packets));
      if (!state->packets)
	{
	  werror("Virtual memory exhausted.\n");
	  abort();
	}
    }
}

static struct hop_info *
hop_info_alloc(struct packet_state *state)
{
  struct hop_info *info;
  
  if (!state->hops)
    {
      /* Initial allocation. */
      state->hop_alloc = 5;
      state->hops = malloc(state->hop_alloc * sizeof(*state->hops));
      state->nhops = 0;
    }
  else
    {
      state->nhops++;
      if (state->nhops + 1 >= state->hop_alloc)
	{
	  state->hop_alloc = state->hop_alloc * 2 + 10;
	  state->hops = realloc(state->hops,
				state->hop_alloc * sizeof(*state->hops));
	  if (!state->hops)
	    {
	      werror("Virtual memory exhausted.\n");
	      abort();
	    }
	}
    }
  assert(state->nhops + 1 < state->hop_alloc);

  info = &state->hops[state->nhops];
  info->addr.s_addr = INADDR_ANY;
  info->name = NULL;

  return info;
}

static const char *
hop_name(struct hop_info *hop)
{
  /* NOTE: The name sometimes points to a static string, sometimes to
   * a malloed one. But that doesn't matter as we never free it
   * anyway. */
  if (hop->name)
    return hop->name;
  else if (hop->addr.s_addr == INADDR_ANY)
    hop->name = "unknown";
  else
    {
      struct sockaddr_in sa;
      struct hostent *ha;
      
      sa.sin_addr = hop->addr;
      sa.sin_port = 0;

      ha = gethostbyaddr((char *) &hop->addr, sizeof(hop->addr), AF_INET);
      if (ha)
	hop->name = sprintf_alloc("%s (%s)", ha->h_name, inet_ntoa(hop->addr));
      
      else
	hop->name = strdup(inet_ntoa(hop->addr));
    }
  return hop->name;
}

static struct error_info *
error_info_alloc(struct packet_state *state)
{
  if (state->nerrors >= state->error_alloc)
    {
      state->error_alloc = state->error_alloc * 2 + 10;
      state->errors = realloc(state->errors,
			      state->error_alloc * sizeof(*state->errors));
      if (!state->errors)
	{
	  werror("Virtual memory exhausted.\n");
	  abort();
	}
    }
  return &state->errors[state->nerrors++];
}


/* oop callbacks */
static void *
recv_callback(oop_source *source, int socket, oop_event event, void *s);

static void *
timeout_callback(oop_source *source, struct timeval time, void *s);

static void *
time_job_callback(oop_source *source, struct timeval time, void *s);

static int
send_packet(struct packet_state *state, int reverse, int ttl)
{
  struct packet_contents contents;
  struct packet_info *info;
  unsigned transmit_size;
  int res;

  assert(state->npackets < state->packet_alloc);

  info = &state->packets[state->npackets];

  memset(&contents, 0, sizeof(contents));
  contents.header[0] = ECHO_PROTOCOL_VERSION;

  if (reverse)
    {
      transmit_size = CONTENTS_SIZE_SMALL;
      contents.header[1] = CONTENTS_ECHO_LARGE;
      contents.header[2] = ttl;
      ttl = DEFAULT_TTL;
    }
  else
    {
      transmit_size = CONTENTS_SIZE_LARGE;
      contents.header[1] = CONTENTS_ECHO_SMALL;
      contents.header[2] = DEFAULT_TTL;
    }
  
  contents.i = state->npackets++;

  if (setsockopt(state->socket, SOL_IP, IP_TTL,
		 &ttl, sizeof(ttl)) < 0)
    werror("setsockopt(SOL_IP, IP_TTL, %d) failed: %s\n",
	    ttl, strerror(errno));
  
  info->ttl = ttl;
  
  do
    {
      if (gettimeofday(&contents.stamp, NULL) < 0)
	{
	  perror("gettimeofday failed");
	  exit(EXIT_FAILURE);
	}
      res = send(state->socket, &contents,
		 transmit_size, 0);
    }
  while (res < 0 && errno == EINTR);
  
  info->time_sent = contents.stamp;
  if (res < 0)
    {
      werror("Sending packet %d failed: %s\n",
	     contents.i, strerror(errno));
      
      info->state = SEND_ERROR;
      info->err = errno;

      return 0;
    }
  else
    {
      info->state = SENT;
      state->to_recv++;
      return 1;
    }
}

static void *
time_job_callback(oop_source *source, struct timeval time, void *s)
{
  struct packet_state *state = (struct packet_state *) s;

  if (time.tv_sec == 0 && time.tv_usec == 0)
    if (gettimeofday(&time, NULL) < 0)
      {
	perror("gettimeofday failed");
	exit(EXIT_FAILURE);
      }
  
  if (state->current(state->ctx, state))
    {
      /* Job finished */
      state->current = NULL;

      time.tv_sec += DEFAULT_TIMEOUT;
      state->timeout = time;

      source->on_time(source, state->timeout, timeout_callback, state);
    }
  else
    {
      unsigned long usec;
      usec = time.tv_usec + state->interval % 1000000;
      time.tv_usec = usec % 1000000;
      time.tv_sec += usec / 1000000;
      
      source->on_time(source, time, time_job_callback, state);
    }
  
  return OOP_CONTINUE;
}

static int
recv_error(int socket, struct packet_state *state)
{
#if IP_RECVERR
  char packet[CONTENTS_SIZE_LARGE];
  struct iovec io = { packet, sizeof(packet) };
  struct msghdr hdr = { 0 };
  char buf[CMSG_SPACE(sizeof(struct sock_extended_err)
		      +sizeof(struct sockaddr_in))];
  int res;
  
  hdr.msg_name = NULL;
  hdr.msg_namelen = 0;
  hdr.msg_iov = &io;
  hdr.msg_iovlen = 1;
  hdr.msg_control = buf;
  hdr.msg_controllen = sizeof(buf);
  hdr.msg_flags = MSG_ERRQUEUE;
      
  res = recvmsg(socket, &hdr, MSG_ERRQUEUE);

  if (res >= 0)
    {
      struct cmsghdr *msg;

      for (msg = CMSG_FIRSTHDR(&hdr);
	   msg;
	   msg = CMSG_NXTHDR(&hdr, msg))
	{
	  if (msg->cmsg_level == SOL_IP
	      && msg->cmsg_type == IP_RECVERR)
	    {
	      struct sock_extended_err *err
		= (struct sock_extended_err *) CMSG_DATA(msg);
	      struct sockaddr *offender = SO_EE_OFFENDER(err);
	      
	      struct error_info *info = error_info_alloc(state);

	      if (offender->sa_family == AF_INET)
		info->where = ((struct sockaddr_in *) offender)->sin_addr;
	      else
		info->where.s_addr = INADDR_ANY;

	      if (err->ee_origin == SO_EE_ORIGIN_ICMP
		  && err->ee_type == ICMP_TIME_EXCEEDED)
		{
		  info->type = ERR_TIME_EXCEEDED;
#if 0		  
		  verbose("Received ICMP_TIME_EXCEEDED from %s\n",
			  inet_ntoa(info->where));
#endif
		  return 1;
		}
	      else
		info->type = ERR_OTHER;
#if 0
	      werror("recv_error: res = %d\n"
		     "ee_errno = %d, ee_origin = %d, ee_type = %d,\n"
		     "ee_code = %d, ee_info = %d, ee_data = %d\n",
		     res,
		     err->ee_errno, err->ee_origin, err->ee_type,
		     err->ee_code, err->ee_info, err->ee_data);
#endif
	    }
	}
    }
#endif /* IP_RECVERR */
  return 0;
}

static void *
check_receive_count(oop_source *source, struct packet_state *state)
{
  assert(state->to_recv);
  state->to_recv--;
  
  if (!state->to_recv && !state->current)
    {
      source->cancel_time(source, state->timeout, timeout_callback, state);
      source->cancel_fd(source, state->socket, OOP_READ);
      
      return OOP_HALT;
    }
  else
    return OOP_CONTINUE;
}

static void *
recv_callback(oop_source *source, int socket, oop_event event, void *s)
{
  struct packet_state *state = (struct packet_state *) s;
  struct packet_contents contents;
  struct packet_info *info;
  struct timeval now;
  unsigned i;
  int res;

  assert(event == OOP_READ);

  res = recv(socket, &contents, sizeof(contents), 0);

  if (res < 0)
    {
      int err = errno;
      if (errno == EINTR)
	return OOP_CONTINUE;

      if (recv_error(socket, state))
	return check_receive_count(source, state);

      werror("recv failed: %s\n", strerror(err));
      
      return OOP_CONTINUE;
    }

  if (gettimeofday(&now, NULL) < 0)
    {
      perror("gettimeofday failed");
      return OOP_CONTINUE;
    }

  if (res < CONTENTS_SIZE_SMALL)
    {
      werror("Received packet of unexpected length %d\n",
	     res);
      return OOP_CONTINUE;
    }

  i = contents.i;
  if (i >= state->npackets)
    {
      werror("Ignoring packet with unexpected index.\n");
      return OOP_CONTINUE;      
    }

  info = &state->packets[i];
  if (info->time_sent.tv_sec != contents.stamp.tv_sec
      || info->time_sent.tv_usec != contents.stamp.tv_usec)
    {
      werror("Ignoring packet with unexpected timestamp.\n");

      if (info->state != RECEIVED)
	info->state = INVALID;
      return OOP_CONTINUE;      
    }

  switch (info->state)
    {
    case SENT:
      info->time_received = now;
      info->state = RECEIVED;

      return check_receive_count(source, state);
      
    case RECEIVED:
      info->state = DUPLICATE;
      break;

    case DUPLICATE:
      break;

    case LOST:
      werror("Ignoring late packet #%d\n", i);
      break;

    default:
      werror("Ignoring unexpected packet.\n");
    }
  return OOP_CONTINUE;      
}  

static void *
timeout_callback(oop_source *source, struct timeval time, void *s)
{
  struct packet_state *state = (struct packet_state *) s;
  source->cancel_fd(source, state->socket, OOP_READ);

  return OOP_HALT;
}

static void
run(oop_source_sys *sys, struct packet_state *state,
    job_function *job, void *ctx, unsigned interval)
{
  unsigned i;
  
  oop_source *source = oop_sys_source(sys);
  
  state->interval = interval;
  state->current = job;
  state->ctx = ctx;
  
  source->on_fd(source, state->socket, OOP_READ, recv_callback, state);
  source->on_time(source, OOP_TIME_NOW, time_job_callback, state);
  oop_sys_run(sys);

  for (i = 0; i<state->npackets; i++)
    {
      if (state->packets[i].state == SENT)
	state->packets[i].state = LOST;
    }
}

static void
make_socket(struct packet_state *state, const char *target, int port)
{
  struct sockaddr_in addr;
  socklen_t addr_len;
  
  /* First check for numerical ip-number */
  if (!inet_aton(target, &addr.sin_addr))
  {
    struct hostent *hp = gethostbyname(target);
    if (!hp || hp->h_addrtype != AF_INET)
      {
	werror("gethostbyname failed.\n");
	exit(EXIT_FAILURE);
      }

    memcpy(&addr.sin_addr, hp->h_addr, hp->h_length);
  }
  
  addr.sin_family = AF_INET;

  addr.sin_port = htons(port);

  state->target = addr.sin_addr;

  /* FIXME: Use PROTO_UDP instead? */  
  state->socket = socket(PF_INET, SOCK_DGRAM, 0);
  if (state->socket < 0)
    {
      perror("socket() failed");
      exit(EXIT_FAILURE);
    }

  if (connect(state->socket, (struct sockaddr *) &addr, sizeof(addr)) < 0)
    {
      perror("connect() failed");
      exit(EXIT_FAILURE);
    }    

  /* Get local address */
  addr_len = sizeof(addr);
  if (getsockname(state->socket,
		  (struct sockaddr *) &addr, &addr_len)  < 0)
    {
      perror("getsockname() failed");
      exit(EXIT_FAILURE);
    }
  else if (addr_len != sizeof(addr))
    {
      werror("Unexpected socklen %d\n", addr_len);
      exit(EXIT_FAILURE);      
    }
  state->local = addr.sin_addr;
  
  {
    int old = fcntl(state->socket, F_GETFL);
    if (old < 0)
    {
      perror("fcntl(F_GETFL) failed");
      exit(EXIT_FAILURE);
    }    
    if (fcntl(state->socket, F_SETFL, old | O_NONBLOCK) < 0)
    {
      perror("fcntl(F_SETFL) failed");
      exit(EXIT_FAILURE);
    }    
  }
#ifdef IP_RECVERR
  {
    int flag = 1;
    if (setsockopt(state->socket , SOL_IP, IP_RECVERR,
		   &flag, sizeof(flag)) < 0)
      {
	perror("Failed to set RECVERR flag");
      }
  }
#endif
}

struct packet
{
  struct timeval stamp;
  char data[800];
};

double
timeval_diff(const struct timeval *a, const struct timeval *b)
{
  return (b->tv_sec - a->tv_sec) + (b->tv_usec - a->tv_usec) / 1000000.0;
}

struct burst_job
{
  unsigned left;
  int reverse;
};

static job_function burst;

static int
burst(void *c, struct packet_state *state)
{
  struct burst_job *job = (struct burst_job *) c;
  assert(job->left);
  
  send_packet(state, job->reverse, DEFAULT_TTL);

  return !--job->left;
}

static void
send_burst(struct oop_source_sys *sys, struct packet_state *state,
	   int reverse,
	   unsigned n, unsigned long interval)
{
  struct burst_job job;
  
  packet_state_alloc(state, n);
  job.left = n;
  job.reverse = reverse;

  run(sys, state, burst, &job, interval);
}


struct probe_job
{
  unsigned left;
  int reverse;
};

static job_function probe;

static int
probe(void *c, struct packet_state *state)
{
  struct burst_job *job = (struct burst_job *) c;
  assert(job->left);

  send_packet(state, job->reverse, DEFAULT_TTL);
  send_packet(state, job->reverse, DEFAULT_TTL);

  return !--job->left;
}

static void
send_probe(struct oop_source_sys *sys, struct packet_state *state,
	   int reverse,
	   unsigned n, unsigned long interval)
{
  struct probe_job job;
  packet_state_alloc(state, 2*n);

  job.left = n;
  job.reverse = reverse;

  run(sys, state, probe, &job, interval);
}

/* FIXME: One way to improve both performance and reliability might be
 * to interleave the probes for different ttl:s.
 *
 * First send probes with alternating ttl 1 and 2. Next, use ttls
 * 1,2,3,4, and go on doubling the interval for each probe. */
struct trace_job
{
  unsigned left;
  int ttl;
  int reverse;
};

static job_function trace;

/* Number of packets that we expect to lose */
#define TRACE_LOSS 1

static int
trace(void *c, struct packet_state *state)
{
  struct trace_job *job = (struct trace_job *) c;
  unsigned i;
  assert(job->left);

  send_packet(state, job->reverse, DEFAULT_TTL);
  for (i = 0; i<TRACE_LOSS; i++)
    send_packet(state, job->reverse, job->ttl);
  send_packet(state, job->reverse, DEFAULT_TTL);

  return !--job->left;
}

static void
send_trace(struct oop_source_sys *sys, struct packet_state *state,
	   int reverse, int ttl,
	   unsigned n, unsigned long interval)
{
  struct trace_job job;
  packet_state_alloc(state, (2+TRACE_LOSS)*n);

  job.left = n;
  job.reverse = reverse;
  job.ttl = ttl;
  
  run(sys, state, trace, &job, interval);
}
  

struct ping_job
{
  unsigned left;
  unsigned ttl;
};

static job_function ping;

static int
ping(void *c, struct packet_state *state)
{
  struct ping_job *job = (struct ping_job *) c;
  assert(job->left);

  send_packet(state, 0, job->ttl);

  return !--job->left;
}

static void
send_ping(struct oop_source_sys *sys, struct packet_state *state,
	  unsigned ttl)
{
  struct ping_job job;
  job.left = 3;
  job.ttl = ttl;
  
  packet_state_alloc(state, job.left);

  run(sys, state, ping, &job, 100000);  
}


#define TOLERANCE 2

static int
estimate_bandwidth(struct oop_source_sys *sys,
		   struct packet_state *state, int reverse,
		   unsigned n)
{
  double *samples;
  unsigned nsamples;

  double *rtt;
  unsigned nrtt;
  
  unsigned i;
  double mean;
  double sigma;
  
  unsigned start = state->npackets;
  send_probe(sys, state, reverse, n, 100000);

  /* The first packet in a pair is sent at time s1 and received at r1.
   * The second is sent at time s2, very soon after s1, and received
   * at time r2. The difference (r2-s1) - (r1-s1) = r2-r1 corresponds
   * to the time it took to transmit the first packet over the
   * bottle-neck link.
   */

  samples = alloca(n * sizeof(*samples));
  rtt = alloca(n * sizeof(*rtt));
  
  for (i = start, nsamples = 0, nrtt = 0; i+1 < state->npackets; i+=2)
    {
      if (state->packets[i].state == RECEIVED)
	{
	  rtt[nrtt++] = timeval_diff(&state->packets[i].time_sent,
				     &state->packets[i].time_received);

	  if (state->packets[i+1].state == RECEIVED)
	    {
	      double d = timeval_diff(&state->packets[i].time_received,
				      &state->packets[i+1].time_received);
	      if (d > 0.0)
		samples[nsamples++] = d;
	    }
	}
    }

  if (nsamples < 5)
    return 0;

  mean = compute_mean(nrtt, rtt, TOLERANCE, &sigma);
  state->rtt_small = mean;
  state->rtt_small_dev = sigma / state->rtt_small;

  verbose("Unloaded rtt = %f  %.1f %%\n",
	  state->rtt_small, 100 * state->rtt_small_dev);
  
  mean = compute_mean(nsamples, samples, TOLERANCE, &sigma);
  state->bandwidth_fast = PACKET_SIZE / mean;
  state->bandwidth_fast_dev = sigma / mean;
  
  verbose("Bandwidth: ");
  verbose("%f%s", state->bandwidth_fast, "bit/s");
  verbose("  %.1f%%\n", 100 * state->bandwidth_fast_dev);

  return 1;
}

/* Returns the drop rate, and the point where we start to drop
 * packets.
 *
 * Let d_i is the number of packets in the interval [0, i] that
 * are dropped, and adapt the model
 *
 * d_i = m_small + d_small i, i <= d_start,
 *     = m_large + d_large (i - d_start), i > d_start 
 *
 * where the parameters d_rate and d_start are chosen by least squares.
 *
 * Use something more sophisticated than brute force for finding d_start.
 */
static double
compute_drop_rate(unsigned n, const struct packet_info *packets,
		  unsigned *drop_start)
{
  double *x;
  double *y;
  unsigned i;
  unsigned dropped;
  unsigned start;
  
  unsigned best_start;
  double best_small;
  double best_large;
  double best_error;
  
  x = alloca(n * sizeof(*x));
  y = alloca(n * sizeof(*y));
  
  for (i = dropped = 0; i < n; i++)
    {
      if (packets[i].state == LOST)
	dropped++;
      x[i] = i;
      y[i] = dropped;
    }

  /* Perform exhastive search over start */

  /* The error, for a given value of start, is
   *
   *   n-1                  n-1                              n-1               
   *   sum   d[i]^2 - 2rate sum   d[i] (i - start) + rate^2  sum  (i-start)^2
   *   i=0                start+1		           start+1            
   *
   * We ignore the first term which is independent of start and rate,
   * and denote the other two sums by e1 and e2.
   *
   * The optimal rate, given start, is rate = e1 / e2, which gives an error of
   *
   *  sum d[i]^2 - e1^2/e2
   *
   */

  best_start = n;
  best_large = 0.0;
  best_small = 0.0;
  best_error = 0.0;
  
  for (start = 0; start < n; start++)
    {      
      double d_small;
      double m_small;
      double e_small;
      
      double d_large;
      double m_large;
      double e_large;
      double error;

      d_small = compute_line(start, x, y, &m_small, &e_small);
      d_large = compute_line(n - start, x + start, y + start,
			     &m_large, &e_large);

      error = e_small + e_large;
#if 0
      verbose("start = %d: error = %f\n"
	     "  d_small = %f, m_small = %f, e_small = %f\n"
	     "  d_large = %f, m_large = %f, e_large = %f\n",
	     start, error,
	     d_small, m_small, e_small,
	     d_large, m_large, e_large);
#endif
      
      if (error < best_error)
	{
	  best_error = error;
	  best_start = start;
	  best_small = d_small;
	  best_large = d_large;
	}
    }
  verbose("drop_rate: %f%% of first %d packets, "
	 "then %f%% packets of %d dropped.\n",
	 best_small * 100, best_start,
	 best_large * 100, n - best_start);
  
  *drop_start = best_start;
  return best_large;
}

/* Estimate the bandwidth of received traffic. */
static double
compute_bandwidth(unsigned n, struct packet_info *packets)
{
  unsigned received;
  unsigned i;
  const struct timeval *start = NULL;
  const struct timeval *end = NULL;
  
  for (i = received = 0; i<n; i++)
    {
      if (packets[i].state == RECEIVED)
	{
	  if (!received)
	    start = &packets[i].time_received;
	  else
	    end = &packets[i].time_received;

	  received++;
	}
    }
  if (start && end)
    return PACKET_SIZE * (received - 1) / timeval_diff(start, end); 
  else
    return 0.0;
}

static double
compute_delay(unsigned n, struct packet_info *packets)
{
  double *samples;
  unsigned nsamples;
  unsigned i;

  double mean;
  double sigma;
  
  samples = alloca(n * sizeof(*samples));

  for (i = nsamples = 0; i < n; i++)
    {
      if (packets[i].state == RECEIVED)
	samples[nsamples++] = timeval_diff(&packets[i].time_sent,
					   &packets[i].time_received);
    }

  mean = compute_mean(nsamples, samples, TOLERANCE, &sigma);
  verbose("Buffer delay: %f s  %.1f%%\n", mean, 100 * sigma/mean);
  return mean;
}

/* Returns buffer size, in bits */
static double
estimate_buffer_size(struct oop_source_sys *sys,
		     struct packet_state *state, int reverse)
{
  double rate;
  unsigned length;
  
  /* We aim for a sending rate of 1.1 times the bandwidth, in order to
   * not fill more than one queue. */
  rate = state->bandwidth_fast;

  /* Start with one megabit data */
  length = (1L<<20) / PACKET_SIZE;

  for (;;)
    {
      double drop_rate;
      unsigned drop_start;
      unsigned interval = 1000000 * (PACKET_SIZE / rate);
      unsigned real_length;
      double bandwidth;
      double ratio;
      unsigned start = state->npackets;
      
      verbose("Sending burst of %d packets, interval %d us\n",
	     length, interval);
      send_burst(sys, state, reverse, length, interval);

      real_length = state->npackets - start;
      /* Find out when we start dropping packets */
      drop_rate = compute_drop_rate(real_length,
				    &state->packets[start],
				    &drop_start);

      bandwidth = compute_bandwidth(real_length, &state->packets[start]);
      verbose("Bandwidth: ");
      verbose("%f%s", bandwidth, "bit/s");
      verbose("\n");

      if (drop_rate > 0.08 && (length - drop_start) > 30)
	{
	  state->bandwidth = bandwidth;
#if 0
	  verbose("Bandwidth, estimated from drop rate: ");
	  verbose("%f%s", rate * (1.0 - drop_rate) , "bit/s");
	  verbose("\n");
#endif
	  state->buffer_drops = (rate - bandwidth)
	    * timeval_diff(&state->packets[start].time_sent,
			   &state->packets[start + drop_start].time_sent) / 8;

	  verbose("Buffer size, estimated from drops: ");
	  verbose("%f%s", state->buffer_drops, "bytes");
	  verbose("\n");
	  
	  state->rtt_large
	    = compute_delay(real_length - drop_start,
			    &state->packets[start + drop_start]);

	  verbose("Delay: %f s\n", state->rtt_large);

	  state->buffer_delay = (state->rtt_large - state->rtt_small)
	    * state->bandwidth / 8;
	  
	  verbose("Buffer size, estimated from delay: ");
	  verbose("%f%s", state->buffer_delay, "bytes");
	  verbose("\n");
		 
	  return 1;
	}

      /* Give up if there were too few packets to estimate the bandwidth. */
      if (bandwidth < 1.0)
	return 0;
	  
      ratio = rate / bandwidth;
      
      if (ratio < 1.1)
	rate = bandwidth * 1.1;
      else
	{
	  if (ratio > 1.5)
	    rate = bandwidth * 1.2;
	  length *= 2;
	}
    }
}

/* Analyzes n * (2 + TRACE_LOSS) packets. Returns 1 if the ttl was
 * large enough for even the small-ttl packets to make it. */
static int
compute_hop_info(struct hop_info *info,
		 unsigned n, struct packet_info *packets)
{
  /* Number of small ttl packets that were received */
  unsigned unexpected_recv = 0;
  /* Number of default ttl packets that were lost */
  unsigned unexpected_lost = 0;
  unsigned i;
  
  double *samples;
  unsigned nsamples;

  double mean;
  double sigma;
  
  samples = alloca(n * sizeof(*samples));
  
  for (i = 0, nsamples = 0; i < n * (2 + TRACE_LOSS); i += (2 + TRACE_LOSS))
    {
      unsigned lost;
      unsigned recv;
      unsigned j;
	  
      lost = ( (packets[i].state != RECEIVED)
	       + (packets[i + 1 + TRACE_LOSS].state != RECEIVED) );
      
      for (j = recv = 0; j<TRACE_LOSS; j++)
	recv += (packets[i + j + 1].state == RECEIVED);

      unexpected_lost += lost;
      unexpected_recv += recv;
	  
      if (!lost)
	{
	  double d = timeval_diff(&packets[i].time_received,
				  &packets[i+ 1+ TRACE_LOSS].time_received);

	  if (d > 0.0)
	    samples[nsamples++] = d;
	}
    }

  if (nsamples)
    {
      mean = compute_mean(nsamples, samples, TOLERANCE, &sigma);
  
      /* The delay corresponds to two packets, one of which was dropped */
      info->bandwidth = PACKET_SIZE * 2 / mean;
      info->bandwidth_dev = sigma / mean;
    }
  else
    info->bandwidth = info->bandwidth_dev = 0.0;

  if (unexpected_recv || unexpected_lost)
    verbose("Received %d small-ttl packets. Lost %d default-ttl packets\n",
	    unexpected_recv, unexpected_lost);
  
  /* We consider ourselves successful if at most one packet with a
   * "small" ttl was received, and at least one default-ttl packet. */
  return (unexpected_lost < 2*n && unexpected_recv < 2);
}

static unsigned
compute_bottleneck(unsigned n, const struct hop_info *hops)
{
  /* Consider the bandwidth b(ttl) as a function of the ttl, 0 < ttl
   * <= n. This function is decreasing, and there should be a constant
   * l such that b(k) ~ b(n-1) = B for k >= l. The smallest such l is
   * the number of hops to the router just after the bottleneck link,
   * and that's what we return.
   *
   * We consider b(k) ~ B iff b(k) < B + TOLERANCE stddev(B).
   */

  /* FIXME: If the input isn't decreasing, ignore offending values. */
  double threshold;
  
  assert(n);

  threshold = hops[n].bandwidth * (1 + TOLERANCE * hops[n].bandwidth_dev);

  for ( ; n > 0; n--)
    {
      /* Skip hops where we have no data.
       *
       * I don't know why this happens but some routers seem to
       * consistently drop the third, default-ttl, packet of our
       * probes. */
      if (hops[n-1].bandwidth == 0.0)
	continue;
      
      if (hops[n-1].bandwidth > threshold)
	return n;
    }
  /* It seeems the first hop is the bottleneck. */
  return 1;
}

static int
get_error_address(struct in_addr *error_addr,
		  unsigned nerrors,
		  const struct error_info *errors)
{
  /* Get the ip. If there are several, display a warning and
   * get the most common. */

  unsigned i;
  struct in_addr addr;
  unsigned nhandled;
  int *handled;

  if (!nerrors)
    return 0;
  
  handled = alloca(nerrors * sizeof(int));
  
  handled[0] = 1;
  addr = errors[0].where; 

  /* FIXME: We don't really want to handle errors of any type but
   * ERR_TIME_EXCEEDED. */
	
  for (nhandled = i = 1; i<nerrors; i++)
    {
      if (errors[i].where.s_addr == addr.s_addr)
	{
	  handled[i] = 1;
	  nhandled++;
	}
      else
	handled[i] = 0;
    }

  if (nhandled < nerrors)
    {
      unsigned best_matched = nhandled;
      werror("Received errors from several sources\n");

      /* NOTE: The probability that this code is executed is pretty
       * low. */
      while (best_matched + nhandled < nerrors)
	{
	  unsigned start;
	  unsigned matched;

	  for (start = 0; start<nerrors && !handled[start]; start++)
	    ;
	  assert(start < nerrors);

	  for (i = start + 1, matched = 1; i<nerrors; i++)
	    if (errors[i].where.s_addr == errors[start].where.s_addr)
	      {
		assert(!handled[i]);
		handled[i] = 1;
		nhandled++;
		matched++;
	      }

	  if (matched > best_matched)
	    {
	      addr = errors[start].where;
	      best_matched = matched;
	    }
	}
    }
  verbose("Errors received from %s\n", inet_ntoa(addr));
  *error_addr = addr;

  return 1;
}

/* Returns number of hops */
static unsigned
estimate_path(struct oop_source_sys *sys,
	      struct packet_state *state, int reverse,
	      unsigned n)
{
  unsigned ttl;
  int goon;

  /* We first collect the hops counting from the sender, and reverse
   * the order later, if doing a reverse trace. */
  {
    struct hop_info *info = hop_info_alloc(state);
    info->addr = reverse ? state->target : state->local;
    info->bandwidth = info->bandwidth_dev = 0.0;
  }
  
  /* We can't send packets with ttl = 0 */
  for (ttl = 1, goon = 1; goon; ttl++)
    {
      unsigned start = state->npackets;
      unsigned error_start = state->nerrors;
      struct hop_info *info;
      
      verbose("Sending probe with ttl = %d\n", ttl);
      
      send_trace(sys, state, reverse, ttl, n, 200000);

      info = hop_info_alloc(state);
      goon = compute_hop_info(info,
			      n, &state->packets[start]);

      if (!get_error_address(&info->addr,
			     state->nerrors - error_start,
			     &state->errors[error_start]))
	info->addr.s_addr = INADDR_ANY;
      
      verbose("tracing: ttl = %d, bandwidth = ", ttl);
      verbose("%f%s", info->bandwidth, "bit/s");
      verbose("  %.1f%%\n", 100 * info->bandwidth_dev);
    }

  assert(state->nhops);
  
  state->bottleneck = compute_bottleneck(state->nhops, state->hops);

  assert(state->bottleneck);
  
  if (!reverse)
    state->hops[state->nhops].addr = state->target;
  else
    {
      unsigned hops = 0;
      unsigned i, j;
      
      state->hops[state->nhops].addr = state->local;
      state->bottleneck = state->nhops - state->bottleneck + 1;
      
      /* Reverse order */
      for (i = 0, j = state->nhops; i<j; i++, j--)
	{
	  struct hop_info tmp = state->hops[i];
	  state->hops[i] = state->hops[j];
	  state->hops[j] = tmp;
	}
      
      /* We need to send trace packets in the forward direction too,
       * in order to get any ICMP_TIME_EXCEEDED messages. */

      for (ttl = 1; ttl <= state->nhops; ttl++)
	{
	  unsigned start = state->npackets;
	  unsigned error_start = state->nerrors;
	  unsigned j;
	  
	  send_ping(sys, state, ttl);

	  for (j = start; j < state->npackets; j++)
	    if (state->packets[j].state == RECEIVED)
	      {
		hops = ttl;
		/* Break out of both loops. */
		goto path_done;
	      }

	  /* Ignore errors, addresses are already initialized to
	   * INADDR_ANY. */
	  get_error_address(&state->hops[ttl].addr,
			    state->nerrors - error_start,
			    &state->errors[error_start]);
	}

    path_done:
      
      if (hops == state->nhops)
	verbose("Same number of hops in forward and reverse paths.\n"
		"Using the forward path to map the reverse path.\n");
      else
	{
	  if (!hops)
	    werror("%d hops on the reverse path, forward path is longer.\n",
		   state->nhops);
	  else
	    werror("%d hops on the forward path, %d on the reverse path.\n",
		   state->nhops, hops);

	  werror("Some or all of the addresses assigned to hops are wrong.\n");
	}
    }

  if (verbose_flag)
    for (ttl = 0; ttl <= state->nhops; ttl++)
      verbose("hop %d: %s\n", ttl, hop_name(&state->hops[ttl]));
  
  return state->bottleneck;
}

static void
report_int(const char *label,
	   int value,
	   const char *extra)
{
  printf("%18s: %d %s\n", label, value, extra); 
}

static void
report_value(const char *label,
	     double value, const char *unit,
	     const char *extra)
{
  printf("%18s: ", label); 
  print_pretty(stdout, value, unit); 
  printf(" %s\n", extra); 
}

static void
report_ratio(const char *label,
	     double value, 
	     const char *extra)
{
  printf("%18s: %.1f %s\n", label, value, extra); 
}

static void
report_ratio_dev(const char *label,
		 double value, double dev, 
		 const char *extra)
{
  printf("%18s: %.1f  %.1f%% %s\n", label, value, 100 * dev, extra); 
}

static void
report_value_dev(const char *label,
		 double value, const char *unit,
		 double dev,
		 const char *extra)
{
  printf("%18s: ", label); 
  print_pretty(stdout, value, unit); 
  printf("  %.1f%% %s\n", 100 * dev, extra); 
}

static void
report_link(const char *label,
	    int reverse,
	    struct hop_info *hop)
{
  printf("%18s: %s %s %s\n", label,
	 hop_name(hop),
	 reverse ? "<-" : "->",
	 hop_name(hop + 1));
}
	    

#if 0
static void
report_value_interval(const char *label,
		      double a, double b, const char *unit,
		      const char *extra)
{
  double min, max;
  if (a < b)
    { min = a; max = b; }
  else
    { min = b; max = a; }

  printf("%s", label);
  print_pretty(stdout, min, unit);
  printf(" - ");
  print_pretty(stdout, max, unit);
  printf(" %s\n", extra);
}
#endif

static void
report(struct packet_state *state, const struct options *options)
{
  double bdp;
  time_t now;
  struct tm *date;

  now = time(NULL);
  date = gmtime(&now);
  
  printf("net-probe report\n"
	 "----------------\n\n");

  printf("Local:     %s\n", hop_name(&state->hops[0]));
  printf("Target:    %s\n", hop_name(&state->hops[state->nhops]));
  if (options->reverse)
    printf("Direction: Reverse\n");
  
  printf("Date:      %d-%02d-%02d %d:%02d:%02d (UTC)\n\n",
	 1900 + date->tm_year, 1 + date->tm_mon, date->tm_mday,
	 date->tm_hour, date->tm_min, date->tm_sec);
  
  report_int("Number of hops", state->nhops, "");

  if (options->long_report)
    {
      /* We measure bandwidths only up to twice the bottleneck bandwidth */
      double threshold = (1 + TRACE_LOSS) * state->bandwidth_fast;
      unsigned i;
      
      printf("\n     Bandwidth                Node\n");
      printf(" %s  -------------------------------\n",
	     options->reverse ? "/\\" : "||");

      for (i = 0; i <= state->nhops; i++)
	{
	  double bandwidth;
	  double dev;
	  
	  printf(" ||                           %s\n",
		 hop_name(&state->hops[i]));

	  if (i == state->nhops)
	    break;

	  bandwidth = state->hops[i + !options->reverse].bandwidth;
	  dev = state->hops[i + !options->reverse].bandwidth_dev;

	  if (i + 1 < state->bottleneck)
	    {
	      if (bandwidth == 0.0)
		printf("unknown");
	      else if (bandwidth * (1 + TOLERANCE * dev) < threshold)
		{
		  printf(" ||  = ");
		  print_pretty(stdout, bandwidth, "bit/s");
		  printf("  %.1f%%", 100 * dev);
		}
	      else
		{
		  printf(" ||  > ");
		  print_pretty(stdout, threshold, "bit/s");
		}
	      printf("\n");
	    }
	  else if (i + 1 == state->bottleneck)
	    {
	      printf(" )(  ");
	      if (bandwidth == 0.0)
		printf("unknown");
	      else
		{
		  printf("= ");
		  print_pretty(stdout, bandwidth, "bit/s");
		  printf("  %.1f%%", 100 * dev);
		}
	      printf(" (bottleneck)\n");
	    }
	  else
	    {
	      printf(" ||  ");
	      if (bandwidth == 0.0)
		printf("unknown");
	      else
		{
		  printf("> ");
		  print_pretty(stdout, bandwidth, "bit/s");
		}
	      printf("\n");
	    }
	}
      printf(" %s\n", options->reverse ? "||" : "\\/");
    }

  printf("\n");
  report_value_dev("Bandwidth estimate",
		   state->bandwidth_fast, "bit/s",
		   state->bandwidth_fast_dev, "(unloaded network)");

  report_value_dev("RTT",
		   state->rtt_small, "s", state->rtt_small_dev,
		   "(unloaded network)");

  bdp = state->rtt_small * state->bandwidth_fast / 8;
  report_value_dev("Bandwidth  delay", bdp, "bytes",
		   state->rtt_small_dev + state->bandwidth_fast_dev,
		   "(unloaded network)");

  report_ratio_dev("alpha", bdp / 1500,
		   state->rtt_small_dev + state->bandwidth_fast_dev,
		   "(for an MTU of 1500 bytes)");

  if (state->bottleneck)
    {
      report_int("Hops to bottleneck", state->bottleneck - 1,
		 "(distance to router on our side of the bottleneck link)");

      report_link("Bottleneck", options->reverse,
		  &state->hops[state->bottleneck - 1]);
    }
  
  if (options->full)
    {
      printf("\n");
      report_value("Bandwidth estimate",
		   state->bandwidth, "bit/s", "(loaded network)");
      report_value("RTT", state->rtt_large, "s", "(loaded network)\n");
      report_value("Bandwidth  delay",
		   state->rtt_large * state->bandwidth / 8, "bytes",
		   "(loaded network)");
      bdp = state->rtt_small * state->bandwidth / 8;
      report_value("Bandwidth  delay", bdp, "bytes",
		   "(unloaded network, from above bandwidth)");
      
      report_ratio("alpha", bdp/1500, "(for an MTU of 1500 bytes)");

      report_value("Buffer size",
		   state->buffer_drops, "bytes", "(estimated from drops)");
      report_value("Buffer size",
		   state->buffer_delay, "bytes", "(estimated from delay)");
      report_value("Queue delay",
		   state->rtt_large - state->rtt_small, "s",
		   "(when queue is full)");
    }
  printf("\n");
  if (bdp < 8.0 * 1500)
    {
      double r = bdp / (8.0 * 1500 - bdp); /* alpha / (8 - alpha) */
      
      report_value("Recommended buffer", bdp * r,
		   "bytes",
		   "(for optimal tcp delay and throughput)");
      report_value("Max queue delay", state->rtt_small * r, "s",
		   "(corresponding to the above buffer size)");
    }
  else
    printf("Bandwidth delay product too large for optimal TCP\n");
  
  printf("\n");
  report_int("Total packets sent", state->npackets, "");
}

static void
parse_options(struct options *options,
	      int argc, char **argv)
{
  int c;
  
  options->target = NULL;
  /* The echo, udp port 7, service. */
  options->port = 7;
  options->full = 0;
  options->reverse = 0;
  options->long_report = 0;

  while ( (c = getopt(argc, argv, "p:qflrv")) != -1)
    switch (c)
      {
      case 'p':
	options->port = atoi(optarg);
	if (options->port <= 0)
	  {
	    werror("Invalid port.\n");
	    exit(EXIT_FAILURE);
	  }
	break;
      case 'q':
	quiet_flag = 1;
	break;
      case 'v':
	verbose_flag = 1;
	break;
      case 'f':
	options->full = 1;
	break;
      case 'l':
	options->long_report = 1;
	break;
      case 'r':
	options->reverse = 1;
	break;
      case '?':
	if (isprint (optopt))
	  werror("Unknown option `-%c'.\n", optopt);
	else
	  werror("Unknown option character `\\x%x'.\n",
		 optopt);
	exit(EXIT_FAILURE);
      default:
	abort();
      }
  
  argc -= optind; argv += optind;
  
  if (!argv[0])
    {
      werror("Usage: net-probe [-p PORT] [-q] [-v] [-f] [-l] [-r] TARGET\n");
      exit(EXIT_FAILURE);
    }
  
  options->target = argv[0];
}

int
main(int argc, char **argv)
{
  int status = EXIT_SUCCESS;
  struct options options;
  
  /* Maybe too large for stack allocation */
  static struct packet_state state;
  struct oop_source_sys *sys;

  parse_options(&options, argc, argv);
  packet_state_init(&state);
  make_socket(&state, options.target, options.port);
  
  sys = oop_sys_new();

  if (!estimate_bandwidth(sys, &state, options.reverse, 50))
    {
      werror("Too few packets made it.\n"
	     "Make sure that the target has the standard echo service enabled,\n"
	     "or run udp-echo on a port of your own choice.\n");
      return EXIT_FAILURE;
    }

  estimate_path(sys, &state, options.reverse, 100);
  
  if (options.full && !estimate_buffer_size(sys, &state, options.reverse))
    {
      werror("Full probe failed.\n");
      options.full = 0;
      status = EXIT_FAILURE;
    }

  oop_sys_delete(sys);
  
  report(&state, &options);
  
  return status;
}
