/* ping-time.c */

/* Copyright (C) 2003 Niels Mller
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <assert.h>
#include <ctype.h>
#include <errno.h>
#include <math.h>
#include <stdarg.h>
#include <stddef.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <time.h>

#include <unistd.h>

#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>

#include <fcntl.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>

#include <netinet/ip_icmp.h>

/* Use Dan Egnor's liboop */
#include <oop.h>

/* Number of seconds to wait for the network to return all our packets. */
#define DEFAULT_TIMEOUT 10

static int quiet_flag = 0;
static int verbose_flag = 0;


static void
werror(const char *format, ...)
     __attribute__((__format__ (__printf__,1, 2)));

static void
verbose(const char *format, ...)
     __attribute__((__format__ (__printf__,1, 2)));

uint16_t
in_checksum(const void *data, unsigned length);

static double
timeval_diff(const struct timeval *a, const struct timeval *b);

static void *
recv_callback(oop_source *source, int socket, oop_event event, void *s);

static void *
time_callback(oop_source *source, struct timeval time, void *s);

static void *
timeout_callback(oop_source *source, struct timeval time, void *s);


struct options
{
  const char *target;
  int interval; /* Milliseconds */
  int count;
};

struct packet_info
{
  enum { PING_FREE = 0, PING_SENT, PING_RECEIVED } state;
  
  struct timeval time_sent;
  /* Long counter, low 16 bits stored in the packets */
  unsigned long index;
};

/* Output formats */
static void
print_rtt(struct packet_info *info, double rtt)
{
  printf("%4lu %ld.%06lu %5f ms\n",
	 info->index, info->time_sent.tv_sec, info->time_sent.tv_usec,
	 rtt * 1000);
}

static void
print_duplicate(struct packet_info *info, double rtt)
{
  if (!quiet_flag)
    printf("%4lu %lu.%06lu %5f ms (DUPLICATE)\n",
	   info->index, info->time_sent.tv_sec, info->time_sent.tv_usec,
	   rtt / 1000);
}

static void
print_lost(struct packet_info *info)
{
  if (!quiet_flag)
    printf("%4lu %lu.%06lu LOST\n",
	   info->index, info->time_sent.tv_sec, info->time_sent.tv_usec);
}

static void
print_unexpected(unsigned long index, const struct timeval *now)
{
  if (!quiet_flag)
    printf("%4lu %lu.%06lu UNEXPECTED\n",
	   index, now->tv_sec, now->tv_usec);
}

/* On the wire format. Read and written at the same end, so it doesn't
   matter if the layout is system dependent. */
struct packet_contents
{
  struct icmphdr header;
  unsigned long index;
};

/* Number of old packets to remember */
#define NPACKETS 0x2000

struct ping_state
{
  int socket;
  struct sockaddr_in target;

  /* Packets left to send */
  unsigned long count;
  struct timeval interval;

  /* Final timeout */
  struct timeval timeout;
  
  /* Number of packets with state PING_SENT */
  unsigned long pending;
  
  /* Id, constructed from our pid */
  uint16_t id;
  
  /* Next sequence number. */
  unsigned long next;

  struct packet_info packet[NPACKETS];
};

static void
ping_state_init(struct ping_state *state, const char *target)
{
  memset(state, 0, sizeof(*state));
  state->id = getpid();
  
  /* First check for numerical ip-number */
  if (!inet_aton(target, &state->target.sin_addr))
  {
    struct hostent *hp = gethostbyname(target);
    if (!hp || hp->h_addrtype != AF_INET)
      {
	werror("gethostbyname failed.\n");
	exit(EXIT_FAILURE);
      }

    memcpy(&state->target.sin_addr, hp->h_addr, hp->h_length);
  }
  
  state->target.sin_family = AF_INET;

  state->socket = socket(AF_INET, SOCK_RAW, IPPROTO_ICMP);
  if (socket < 0)
    {
      werror("Creating socket failed: %s.\n", strerror(errno));
      exit(EXIT_FAILURE);
    }
}

static void
send_ping(struct ping_state *state)
{
  unsigned long index;
  struct packet_info *info;
  int res;
  
  struct packet_contents packet;

  index = state->next++;
  info = state->packet + (index % NPACKETS);
  if (info->state == PING_SENT)
    {
      assert(state->pending);
      print_lost(info);
      info->state = PING_FREE;
      state->pending--;
    }
  
  info->index = index;

  /* Construct packet */
  memset(&packet, 0, sizeof(packet));
  packet.header.type = ICMP_ECHO;
  packet.header.un.echo.id = state->id;
  packet.header.un.echo.sequence = index & 0xffff;
  packet.index = index;

  packet.header.checksum = in_checksum(&packet, sizeof(packet));

  do {
    gettimeofday(&info->time_sent, NULL);
    res = sendto(state->socket, (char *) & packet, sizeof(packet), 0,
		 (struct sockaddr *) &state->target, sizeof(state->target));
  } while (res < 0 && errno == EINTR);

  if (res < 0)
    {
      info->state = PING_FREE;
      werror("sendto failed for packet %lu: %s.\n",
	     info->index, strerror(errno));
    }
  else if (res != sizeof(packet))
    {
      info->state = PING_FREE;
      werror("sendto truncated packet %lu from %zu to %d octets.\n",
	     info->index, sizeof(packet), res);
    }
  else
    {
      info->state = PING_SENT;
      state->pending++;
    }
  }

static void
recv_ping(struct ping_state *state)
{
  char packet[100];
  int res;
  struct timeval now;
  unsigned ip_length;
  struct packet_contents *data;
  struct packet_info *info;
  double rtt;
  
  struct sockaddr_in from;
  socklen_t from_length;

  do
    {
      from_length = sizeof(from);
      res = recvfrom(state->socket, packet, sizeof(packet), 0,
		     (struct sockaddr *) &from, &from_length);
    }
  while (res < 0 && errno == EINTR);

  gettimeofday(&now, NULL);

  if (from_length != sizeof(from)
      || from.sin_family != state->target.sin_family)
    verbose("Ignoring packet with unexpected source address size.\n");

  if (memcmp(&from.sin_addr, &state->target.sin_addr, sizeof(from.sin_addr))!= 0)
    {
      verbose("Ignoring packet with unexpected source address %s:%d.\n",
	      inet_ntoa(from.sin_addr), ntohs(from.sin_port));
      return;
    }
  ip_length = ((struct iphdr *) packet)->ihl << 2;
  if (res != ip_length + sizeof(*data))
    {
      verbose("Received icmp packet of length %d, expected %zu\n",
	      res, ip_length + sizeof(*data));
      return;
    }
  data = (struct packet_contents *) (packet + ip_length);

  /* Does the kernel check the checksum for us? */
    
  if (data->header.type != ICMP_ECHOREPLY)
    {
      verbose("Ignoring icmp packet with unexpected type %d.\n",
	      data->header.type);
      return;
    }
  if (data->header.un.echo.id != state->id)
    {
      verbose("Ignoring icmp packet with unexpected id %d.\n",
	      data->header.un.echo.id);
      return;
    }

  info = state->packet + data->header.un.echo.sequence % NPACKETS;

  if (info->state == PING_FREE || info->index != data->index)
    {
      print_unexpected(data->index, &now);
      return;
    }
  rtt = timeval_diff(&info->time_sent, &now);

  if (info->state == PING_RECEIVED)
    {
      print_duplicate(info, rtt);
    }
  else
    {
      assert(state->pending);
      print_rtt(info, rtt);
      info->state = PING_RECEIVED;
      state->pending--;
    }
}


/* OOP callbacks */  
static void *
recv_callback(oop_source *source, int socket, oop_event event, void *s)
{
  struct ping_state *state = (struct ping_state *) s;
  assert(event == OOP_READ);
  
  recv_ping(state);

  if (state->count == 0 && state->pending == 0)
    {
      source->cancel_fd(source, state->socket, OOP_READ);
      source->cancel_time(source, state->timeout, timeout_callback, state);
      return OOP_HALT;
    }
      
  return OOP_CONTINUE;      
} 

static void *
time_callback(oop_source *source, struct timeval time, void *s)
{
  struct ping_state *state = (struct ping_state *) s;

  assert(state->count);
  
  if (time.tv_sec == 0 && time.tv_usec == 0)
    gettimeofday(&time, NULL);

  send_ping(state);

  state->count--;
  if (state->count)
    {
      /* More to send */
      unsigned long usec;
      usec = time.tv_usec + state->interval.tv_usec;
      time.tv_usec = usec % 1000000;
      time.tv_sec += usec / 1000000 + state->interval.tv_sec;

      source->on_time(source, time, time_callback, state);
    }
  else
    {
      assert(state->pending);
      verbose("Waiting %d seconds for replies.\n", DEFAULT_TIMEOUT);
      
      time.tv_sec += DEFAULT_TIMEOUT;
      state->timeout = time;
      source->on_time(source, state->timeout, timeout_callback, state);
    }
  
  return OOP_CONTINUE;
}

static void *
timeout_callback(oop_source *source, struct timeval time, void *s)
{
  struct ping_state *state = (struct ping_state *) s;
  unsigned i;
  
  for (i = 0; i < NPACKETS; i++)
    {
      if (state->packet[i].state == PING_SENT)
	print_lost(state->packet + i);
    }

  source->cancel_fd(source, state->socket, OOP_READ);  
  return OOP_HALT;
}

static void
run(oop_source_sys *sys, struct ping_state *state,
    unsigned long count, unsigned interval)
{
  oop_source *source = oop_sys_source(sys);

  state->count = count;
  state->interval.tv_sec = interval / 1000;
  state->interval.tv_usec = (interval % 1000) * 1000;
  
  source->on_fd(source, state->socket, OOP_READ, recv_callback, state);
  source->on_time(source, OOP_TIME_NOW, time_callback, state);
  oop_sys_run(sys);
}

static void
parse_options(struct options *options,
	      int argc, char **argv)
{
  int c;
  
  options->target = NULL;
  options->interval = 500;
  options->count = 200;
  
  while ( (c = getopt(argc, argv, "i:c:rv")) != -1)
    switch (c)
      {
      case 'i':
	options->interval = atoi(optarg);
	if (options->interval <= 0)
	  {
	    werror("Invalid sample interval.\n");
	    exit(EXIT_FAILURE);
	  }
	break;
      case 'c':
	options->count = atoi(optarg);
	if (options->count <= 0)
	  {
	    werror("Invalid packet count.\n");
	    exit(EXIT_FAILURE);	    
	  }
      case 'q':
	quiet_flag = 1;
	break;
      case 'v':
	verbose_flag = 1;
	break;
      case '?':
	if (isprint (optopt))
	  werror("Unknown option `-%c'.\n", optopt);
	else
	  werror("Unknown option character `\\x%x'.\n",
		 optopt);
	exit(EXIT_FAILURE);
      default:
	abort();
      }
  
  argc -= optind; argv += optind;
  
  if (!argv[0])
    {
      werror("Usage: ping-time [-i interval (ms)] [-c count] [-q] [-v] TARGET\n");
      exit(EXIT_FAILURE);
    }
  
  options->target = argv[0];
}


int
main(int argc, char **argv)
{
  int status = EXIT_SUCCESS;
  struct options options;
  
  static struct ping_state state;
  struct oop_source_sys *sys;

  parse_options(&options, argc, argv);
  ping_state_init(&state, options.target);
  
  sys = oop_sys_new();
  run(sys, &state, options.count, options.interval);
  
  oop_sys_delete(sys);
  
  return status;
}


/* Miscellaneous helper functions */

/* Checksum computed according to RFC 1071 */
uint16_t
in_checksum(const void *data, unsigned length)
{
  uint32_t sum;
  const uint16_t *p = (const uint16_t *) data;
  
  for (sum = 0; length >= 2; length -= 2)
    sum += *p++;

  if (length)
    sum += *(uint8_t *) p;

  while (sum & ~0xffff)
    sum = (sum & 0xffff) + (sum >> 16);

  return ~sum;
}

static double
timeval_diff(const struct timeval *a, const struct timeval *b)
{
  return (b->tv_sec - a->tv_sec) + (b->tv_usec - a->tv_usec) / 1000000.0;
}

static void
werror(const char *format, ...)
{
  if (!quiet_flag)
    {
      va_list args;
      va_start(args, format);
      vfprintf(stderr, format, args);
      va_end(args);
    }
}

static void
verbose(const char *format, ...)
{
  if (verbose_flag)
    {
      va_list args;
      va_start(args, format);
      vfprintf(stderr, format, args);
      va_end(args);
    }
}
